@@ -1,11 +1,11 @@
 """ mxTextTools - A tools package for fast text processing.
 
     Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
-    Copyright (c) 2000-2001, eGenix.com Software GmbH; mailto:info@egenix.com
+    Copyright (c) 2000-2011, eGenix.com Software GmbH; mailto:info@egenix.com
     See the documentation for further information on copyrights,
     or contact the author. All Rights Reserved.
 """
-import string,types
+import types
 
 #
 # import the C module and the version number
@@ -28,11 +28,13 @@ from Constants.Sets import *
 #
 def format_entry(table,i,
 
-                 TupleType=types.TupleType):
+                 TupleType=types.TupleType,StringTypes=types.StringTypes):
 
     """ Returns a pp-formatted tag table entry as string 
     """
     e = table[i]
+    if type(e) in StringTypes:
+        return '%-15.15s : %s' % ('', e)
     jne = 0
     je = 1
     t,c,m = e[:3]
@@ -48,7 +50,11 @@ def format_entry(table,i,
         m = repr(m)
         if len(m) > 17:
             m = m[:17]+'...'
-    return '%-15.15s : %-30s : jne=%+i : je=%+i' % \
+    if type(jne) not in StringTypes:
+        jne = '%+i' % jne
+    if type(je) not in StringTypes:
+        je = '%+i' % je
+    return '%-15.15s : %-30s : jne=%s : je=%s' % \
            (repr(t),'%-.15s : %s'%(c,m),jne,je)
 
 def format_table(table,i=-1):
@@ -61,7 +67,7 @@ def format_table(table,i=-1):
             l.append('--> '+format_entry(table,j))
         else:
             l.append('    '+format_entry(table,j))
-    return string.join(l,'\n')+'\n'
+    return '\n'.join(l)+'\n'
 
 def print_tagtable(table):
 
@@ -149,7 +155,7 @@ def word_in_list(l):
     keys = d.keys()
     if len(keys) < 18: # somewhat arbitrary bound
         # fast hint for small sets
-        t.append((None,IsIn,string.join(d.keys(),'')))
+        t.append((None,IsIn,''.join(d.keys())))
         t.append((None,Skip,-1))
     # test groups
     for c, group in d.items():
@@ -167,32 +173,32 @@ def word_in_list(l):
 # Extra stuff useful in combination with the C functions
 #
 
-def replace(text,what,with,start=0,stop=None,
+def replace(text,match,replacement,start=0,stop=None,
 
-            SearchObject=BMS,join=join,joinlist=joinlist,tag=tag,
-            string_replace=string.replace,type=type,
+            SearchObject=TextSearch,join=join,joinlist=joinlist,tag=tag,
+            type=type,
             StringType=types.StringType):
 
-    """A fast replacement for string.replace.
+    """ A fast replacement for str.replace().
     
-       what can be given as string or search object.
+        match can be given as string or search object.
 
-       This function is a good example for the AppendTagobj-flag usage
-       (the taglist can be used directly as joinlist).
+        This function is a good example for the AppendTagobj-flag
+        usage (the taglist can be used directly as joinlist).
        
     """
-    if type(what) == StringType:
-        so = SearchObject(what)
+    if type(match) is not TextSearchType:
+        so = SearchObject(match)
     else:
-        so = what
-        what = so.match
+        so = match
+        match = so.match
     if stop is None:
-        if start == 0 and len(what) < 2:
-            return string_replace(text,what,with)
+        if start == 0 and len(match) < 2:
+            return text.replace(match,replacement)
         stop = len(text)
     t = ((text,sWordStart,so,+2),
          # Found something, replace and continue searching
-         (with,Skip+AppendTagobj,len(what),-1,-1),
+         (replacement,Skip+AppendTagobj,len(match),-1,-1),
          # Rest of text
          (text,Move,ToEOF)
          )
@@ -203,62 +209,66 @@ def replace(text,what,with,start=0,stop=None,
 
 # Alternative (usually slower) versions using different techniques:
 
-def _replace2(text,what,with,start=0,stop=None,
+def _replace2(text,match,replacement,start=0,stop=None,
 
               join=join,joinlist=joinlist,tag=tag,
-              StringType=types.StringType,BMS=BMS):
+              TextSearchType=TextSearchType,TextSearch=TextSearch):
 
-    """Analogon to string.replace; returns a string with all occurences
-       of what in text[start:stop] replaced by with
-       - uses a one entry tag-table and a Boyer-Moore-Search-object
-       - what can be a string or a BMS/FS search object
-       - it's faster than string.replace in those cases, where
-         the what-string gets long and/or many replacements are found;
-         faster meaning from a few percent up to many times as fast
-       - start and stop define the slice of text to work in
-       - stop defaults to len(text)
+    """ Analogon to str.replace(); returns a string with all occurences
+        of match in text[start:stop] replaced by replacement.
+       
+        This version uses a one entry tag-table and a
+        Boyer-Moore-Search-object.  match can be a string or a
+        TextSearch search object.  It's faster than str.replace() in
+        those cases, where the match-string gets long and/or many
+        replacements are found; faster meaning from a few percent up
+        to many times as fast
+
+        start and stop define the slice of text to work in.  stop
+        defaults to len(text).
+       
     """
     if stop is None:
         stop = len(text)
-    if type(what) == StringType:
-        what=BMS(what)
-    t = ((with,sFindWord,what,+1,+0),)
+    if type(match) is not TextSearchType:
+        match=TextSearch(match)
+    t = ((replacement,sFindWord,match,+1,+0),)
     found,taglist,last = tag(text,t,start,stop)
     if not found: 
         return text
     return join(joinlist(text,taglist))
 
-def _replace3(text,what,with,
+def _replace3(text,match,replacement,
 
-              join=string.join,FS=FS,
-              StringType=types.StringType):
+              TextSearch=TextSearch,
+              TextSearchType=TextSearchType):
 
-    if type(what) == StringType:
-        what=FS(what)
-    slices = what.findall(text)
+    if type(match) is not TextSearchType:
+        match=TextSearch(match)
+    slices = match.findall(text)
     if not slices:
         return text
     l = []
     x = 0
     for left,right in slices:
-        l.append(text[x:left] + with)
+        l.append(text[x:left] + replacement)
         x = right
     l.append(text[x:])
-    return join(l,'')
+    return ''.join(l)
 
-def _replace4(text,what,with,
+def _replace4(text,match,replacement,
 
-              join=join,joinlist=joinlist,tag=tag,FS=FS,
-              StringType=types.StringType):
+              join=join,joinlist=joinlist,tag=tag,TextSearch=TextSearch,
+              TextSearchType=TextSearchType):
 
-    if type(what) == StringType:
-        what=FS(what)
-    slices = what.findall(text)
+    if type(match) is not TextSearchType:
+        match=TextSearch(match)
+    slices = match.findall(text)
     if not slices:
         return text
     repl = [None]*len(slices)
     for i in range(len(slices)):
-        repl[i] = (with,)+slices[i]
+        repl[i] = (replacement,)+slices[i]
     return join(joinlist(text,repl))
 
 def multireplace(text,replacements,start=0,stop=None,
@@ -283,27 +293,27 @@ def multireplace(text,replacements,start=0,stop=None,
     else:
         return join(joinlist(text,replacements,start))
 
-def find(text,what,start=0,stop=None,
+def find(text,match,start=0,stop=None,
 
-         SearchObject=FS):
+         SearchObject=TextSearch):
 
-    """ A faster replacement for string.find().
+    """ A faster replacement for str.find().
 
         Uses a search object for the task. Returns the position of the
-        first occurance of what in text[start:stop]. stop defaults to
+        first occurance of match in text[start:stop]. stop defaults to
         len(text).  Returns -1 in case no occurance was found.
         
     """
     if stop is not None:
-        return SearchObject(what).find(text,start,stop)
+        return SearchObject(match).find(text,start,stop)
     else:
-        return SearchObject(what).find(text,start)
+        return SearchObject(match).find(text,start)
 
-def findall(text,what,start=0,stop=None,
+def findall(text,match,start=0,stop=None,
 
-            SearchObject=FS):
+            SearchObject=TextSearch):
 
-    """ Find all occurances of what in text.
+    """ Find all occurances of match in text.
 
         Uses a search object for the task. Returns a list of slice
         tuples (l,r) marking the all occurances in
@@ -312,23 +322,22 @@ def findall(text,what,start=0,stop=None,
         
     """
     if stop is not None:
-        return SearchObject(what).findall(text,start,stop)
+        return SearchObject(match).findall(text,start,stop)
     else:
-        return SearchObject(what).findall(text,start)
+        return SearchObject(match).findall(text,start)
 
 def split(text,sep,start=0,stop=None,translate=None,
 
-          SearchObject=FS):
+          SearchObject=TextSearch):
 
-    """ A faster replacement for string.split().
+    """ A faster replacement for str.split().
 
         Uses a search object for the task. Returns the result of
         cutting the text[start:stop] string into snippets at every sep
         occurance in form of a list of substrings. translate is passed
-        to the search object as translation string.
+        to the search object as translation string if given.
 
-        XXX convert to a C function... or even better, add as method
-        to search objects.
+        XXX Undocumented.
 
     """
     if translate:
@@ -362,12 +371,15 @@ def tagdict(text,*args):
     """ Tag a text just like the function tag() and then convert
         its output into a dictionary where the tagobjects reference
         their respective strings
-        - this function emulates the interface of tag()
-        - in contrast to tag() this funtion *does* make copies
-          of the found stings
-        - returns a tuple (rc,tagdict,next) with the same meaning
-          of rc and next as tag(); tagdict is the new dictionary - 
-          None in case rc is 0
+
+        This function emulates the interface of tag().  In contrast to
+        tag() this funtion *does* make copies of the found stings,
+        though.
+
+        Returns a tuple (rc,tagdict,next) with the same meaning of rc
+        and next as tag(); tagdict is the new dictionary or None in
+        case rc is 0.
+          
     """
     rc,taglist,next = apply(tag,(text,)+args)
     if not rc:
@@ -378,7 +390,7 @@ def tagdict(text,*args):
         pfx = str(o)
         d[pfx] = text[l:r]
         if s:
-            tagdict(text,dict,pfx+'.',s)
+            tagdict(text,d,pfx+'.',s)
     return (rc,d,next)
 
 def invset(chars):
@@ -389,19 +401,19 @@ def invset(chars):
 
 def is_whitespace(text,start=0,stop=None,
 
-                  nonwhitespace=nonwhitespace_set,setfind=setfind):
+                  charset=nonwhitespace_charset):
 
     """ Return 1 iff text[start:stop] only contains whitespace
         characters (as defined in Constants/Sets.py), 0 otherwise.
+
     """
     if stop is None:
         stop = len(text)
-    i = setfind(text,nonwhitespace,start,stop)
-    return (i < 0)
+    return (charset.search(text, 1, start, stop) is None)
 
 def collapse(text,separator=' ',
 
-             join=join,setsplit=setsplit,collapse_set=set(newline+whitespace)):
+             join=join,charset=CharSet(newline+whitespace)):
 
     """ Eliminates newline characters and compresses whitespace
         characters into one space.
@@ -410,12 +422,14 @@ def collapse(text,separator=' ',
         this function called with '-' separator ;-)
         
     """
-    return join(setsplit(text,collapse_set),separator)
+    return join(charset.split(text), separator)
+
+###
 
 _linesplit_table = (
     (None,Is,'\r',+1),
     (None,Is,'\n',+1),
-    ('line',AllInSet+AppendMatch,set('\r\n',0),+1,-2),
+    ('line',AllInCharSet+AppendMatch,CharSet('^\r\n'),+1,-2),
     (None,EOF,Here,+1,MatchOk),
     ('empty line',Skip+AppendMatch,0,0,-4),
     )
@@ -435,17 +449,19 @@ def splitlines(text,
         and Unix origins in a portable way.
         
     """
-    return tag(text,linesplit_table)[1]
+    return tag(text, linesplit_table)[1]
 
 # Alias for backward compatibility
 linesplit = splitlines
 
+###
+
 _linecount_table = (
     (None,Is,'\r',+1),
     (None,Is,'\n',+1),
-    ('line',AllInSet+AppendTagobj,set('\r\n',0),+1,-2),
+    ('line',AllInCharSet,CharSet('^\r\n'),+1,-2),
     (None,EOF,Here,+1,MatchOk),
-    ('empty line',Skip+AppendTagobj,0,0,-4),
+    ('empty line',Skip,0,0,-4),
     )
 
 def countlines(text,
@@ -456,18 +472,21 @@ def countlines(text,
 
         Line ends are treated just like for splitlines() in a
         portable way.
+        
     """
-    return len(tag(text,linecount_table)[1])
+    return len(tag(text, linecount_table)[1])
+
+###
 
 _wordsplit_table = (
-    (None,AllInSet,whitespace_set,+1),
-    ('word',AllInSet+AppendMatch,nonwhitespace_set,+1,-1),
+    (None,AllInCharSet,whitespace_charset,+1),
+    ('word',AllInCharSet+AppendMatch,nonwhitespace_charset,+1,-1),
     (None,EOF,Here,+1,MatchOk),
     )
 
 def splitwords(text,
 
-               setsplit=setsplit,whitespace_set=whitespace_set):
+               charset=whitespace_charset):
 
     """ Split text into a list of single words.
 
@@ -475,7 +494,76 @@ def splitwords(text,
         before adding the words to the list.
         
     """
-    return setsplit(text,whitespace_set)
+    return charset.split(text)
+
+###
+
+_quoted_split_tagtables = {}
+
+def quoted_split_tagtable(separators, whitespace):
+
+    table = _quoted_split_tagtables.get((separators, whitespace))
+    if not table:
+        target_charset = CharSet('^' + separators + whitespace)
+        whitespace_charset = CharSet(whitespace)
+        table = (
+            'intro',
+            (None, AllInCharSet, whitespace_charset, +1),
+            (None, EOF, Here, +1, MatchOk),
+
+            'detect_target_type',
+            (None, Is, '"', +1, 'double_quoted_target'),
+            (None, Is, '\'', +1, 'single_quoted_target'),
+            (None, IsIn, separators, 'unquoted_target', 'empty_target'),
+
+            'single_quoted_target',
+            ('match', AllNotIn + AppendMatch, '\'', +1, +2),
+            ('match', Skip + AppendMatch, 0),
+            (None, Is, '\'', Fail, 'read_separator'),
+
+            'double_quoted_target',
+            ('match', AllNotIn + AppendMatch, '"', +1, +2),
+            ('match', Skip + AppendMatch, 0),
+            (None, Is, '"', Fail, 'read_separator'),
+
+            'empty_target',
+            ('match', Skip + AppendMatch, 0, +1, 'check_eof'),
+            (None, EOF, Here, +1, MatchOk),
+
+            'unquoted_target',
+            ('match', AllInCharSet + AppendMatch, target_charset,
+             				'empty_target', 'read_separator'),
+
+            'read_separator',
+            (None, AllInCharSet, whitespace_charset, +1),
+            (None, EOF, Here, +1, MatchOk),
+            (None, IsIn, separators, +1),
+            
+            'check_eof',
+            (None, AllInCharSet, whitespace_charset, +1),
+            ('match', EOF + AppendMatch, Here, 'detect_target_type'),
+
+            )
+        _quoted_split_tagtables[(separators, whitespace)] = table
+    return table
+
+def quoted_split(text, separator=' ', whitespace=' \t\r\n'):
+
+    """ Split text into a list of strings using separator as
+        string separator.
+
+        Quoted parts of the text are not scanned for separators.
+        Single and double quotes are allowed. Escaped quotes are not
+        permitted.
+
+        whitespace defines which characters to strip from the strings
+        in the list.
+
+        XXX Undocumented.
+
+    """
+    table = quoted_split_tagtable(separator, whitespace)
+    return tag(text, table)[1]
 
 #
 # Testing and benchmarking
@@ -496,24 +584,32 @@ class _timer:
     atime = 0
 
     def start(self,
+              
               clock=time.clock,time=time.time):
+        
         self.atime = time()
         self.utime = clock()
 
     def stop(self,
+             
              clock=time.clock,time=time.time):
+
         self.utime = clock() - self.utime
         self.atime = time() - self.atime
         return self.utime,self.atime
 
     def usertime(self,
+                 
                  clock=time.clock,time=time.time):
+        
         self.utime = clock() - self.utime
         self.atime = time() - self.atime
         return self.utime
 
     def abstime(self,
+                
                 clock=time.clock,time=time.time):
+        
         self.utime = clock() - self.utime
         self.atime = time() - self.atime
         return self.utime
@@ -544,7 +640,6 @@ def _bench(file='mxTextTools/mxTextTools.c'):
         print
         
     text = open(file).read()
-    import string
 
     t = _timer()
     print 'Working on a %i byte string' % len(text)
@@ -554,16 +649,16 @@ def _bench(file='mxTextTools/mxTextTools.c'):
         print 'Replacing strings'
         print '-'*72
         print
-        for what,with in (('m','M'),('mx','MX'),('mxText','MXTEXT'),
+        for match,replacement in (('m','M'),('mx','MX'),('mxText','MXTEXT'),
                           ('hmm','HMM'),('hmmm','HMM'),('hmhmm','HMM')):
-            print 'Replace "%s" with "%s"' % (what,with)
+            print 'Replace "%s" with "%s"' % (match,replacement)
             t.start()
             for i in range(100):
-                rtext = string.replace(text,what,with)
-            print 'with string.replace:',t.stop(),'sec.'
+                rtext = text.replace(match,replacement)
+            print 'with str.replace():',t.stop(),'sec.'
             t.start()
             for i in range(100):
-                ttext = replace(text,what,with)
+                ttext = replace(text,match,replacement)
             print 'with tag.replace:',t.stop(),'sec.'
             if ttext != rtext:
                 print 'results are NOT ok !'
@@ -571,7 +666,7 @@ def _bench(file='mxTextTools/mxTextTools.c'):
                 mismatch(rtext,ttext)
             t.start()
             for i in range(100):
-                ttext = _replace2(text,what,with)
+                ttext = _replace2(text,match,replacement)
             print 'with tag._replace2:',t.stop(),'sec.'
             if ttext != rtext:
                 print 'results are NOT ok !'
@@ -579,7 +674,7 @@ def _bench(file='mxTextTools/mxTextTools.c'):
                 print rtext
             t.start()
             for i in range(100):
-                ttext = _replace3(text,what,with)
+                ttext = _replace3(text,match,replacement)
             print 'with tag._replace3:',t.stop(),'sec.'
             if ttext != rtext:
                 print 'results are NOT ok !'
@@ -587,7 +682,7 @@ def _bench(file='mxTextTools/mxTextTools.c'):
                 print rtext
             t.start()
             for i in range(100):
-                ttext = _replace4(text,what,with)
+                ttext = _replace4(text,match,replacement)
             print 'with tag._replace4:',t.stop(),'sec.'
             if ttext != rtext:
                 print 'results are NOT ok !'
@@ -601,19 +696,17 @@ def _bench(file='mxTextTools/mxTextTools.c'):
         print '-'*72
         print
 
-        op = string.lower
         t.start()
         for i in range(1000):
-            op(text)
+            text.lower()
         t.stop()
-        print ' string.lower:',t
+        print ' str.lower():',t
 
-        op = string.upper
         t.start()
         for i in range(1000):
-            op(text)
+            text.upper()
         t.stop()
-        print ' string.upper:',t
+        print ' str.upper():',t
 
         op = upper
         t.start()
@@ -630,9 +723,9 @@ def _bench(file='mxTextTools/mxTextTools.c'):
         print ' TextTools.lower:',t
 
         print 'Testing...',
-        ltext = string.lower(text)
+        ltext = text.lower()
         assert ltext == lower(text)
-        utext = string.upper(text)
+        utext = text.upper()
         assert utext == upper(text)
         print 'ok.'
 
@@ -644,12 +737,11 @@ def _bench(file='mxTextTools/mxTextTools.c'):
 
         l = setsplit(text,whitespace_set)
 
-        op = string.join
         t.start()
         for i in range(1000):
-            op(l)
+            ' '.join(l)
         t.stop()
-        print ' string.join:',t
+        print ' str.join():',t
 
         op = join
         t.start()
@@ -658,12 +750,11 @@ def _bench(file='mxTextTools/mxTextTools.c'):
         t.stop()
         print ' TextTools.join:',t
 
-        op = string.join
         t.start()
         for i in range(1000):
-            op(l,' ')
+            '_'.join(l)
         t.stop()
-        print ' string.join with separator:',t
+        print ' str.join() with separator:',t
 
         op = join
         t.start()
@@ -695,12 +786,11 @@ def _bench(file='mxTextTools/mxTextTools.c'):
         print '-'*72
         print
 
-        op = string.split
         t.start()
         for i in range(100):
-            op(text)
+            text.split()
         t.stop()
-        print ' string.split whitespace:',t,'(',len(op(text)),'snippets )'
+        print ' str.split whitespace:',t,'(',len(text.split()),'snippets )'
 
         op = setsplit
         ws = whitespace_set
@@ -710,15 +800,14 @@ def _bench(file='mxTextTools/mxTextTools.c'):
         t.stop()
         print ' TextTools.setsplit whitespace:',t,'(',len(op(text,ws)),'snippets )'
 
-        assert string.split(text) == setsplit(text,ws)
+        assert text.split() == setsplit(text,ws)
 
-        op = string.split
         sep = 'a'
         t.start()
         for i in range(100):
-            op(text,sep)
+            text.split(sep)
         t.stop()
-        print ' string.split at "a":',t,'(',len(op(text,sep)),'snippets )'
+        print ' str.split at "a":',t,'(',len(text.split(sep)),'snippets )'
 
         op = split
         sep = 'a'
@@ -744,15 +833,14 @@ def _bench(file='mxTextTools/mxTextTools.c'):
         t.stop()
         print ' TextTools.setsplit at "a":',t,'(',len(op(text,sep)),'snippets )'
 
-        # Note: string.split and setsplit don't work identically !
+        # Note: str.split and setsplit don't work identically !
 
-        op = string.split
         sep = 'int'
         t.start()
         for i in range(100):
-            op(text,sep)
+            text.split(sep)
         t.stop()
-        print ' string.split at "int":',t,'(',len(op(text,sep)),'snippets )'
+        print ' str.split at "int":',t,'(',len(text.split(sep)),'snippets )'
 
         op = split
         sep = 'int'
@@ -770,13 +858,12 @@ def _bench(file='mxTextTools/mxTextTools.c'):
         t.stop()
         print ' TextTools.setsplit at "i", "n", "t":',t,'(',len(op(text,sep)),'snippets )'
 
-        op = string.split
         sep = 'register'
         t.start()
         for i in range(100):
-            op(text,sep)
+            text.split(sep)
         t.stop()
-        print ' string.split at "register":',t,'(',len(op(text,sep)),'snippets )'
+        print ' str.split at "register":',t,'(',len(text.split(sep)),'snippets )'
 
         op = split
         sep = 'register'
